begin
  require 'mcollective'
rescue LoadError
  module MCollective
    module RPC
      def mcollective_is_missing!
        raise "It appears that MCollective is not installed.  Please install it."
      end

      def rpcclient(*args)
        mcollective_is_missing!
      end

      def custom_request(*args)
        mcollective_is_missing!
      end
    end
  end
end

class Emcee
  class Runner
    include MCollective::RPC
    attr_reader :filter, :nodeset
    DISCOVERY_TIMEOUT = ENV['LM_DISCOVERY_TIMEOUT'] ? ENV['LM_DISCOVERY_TIMEOUT'].to_i : 4
    INVENTORY_RETRIES = ENV['LM_INVENTORY_RETRIES'] ? ENV['LM_INVENTORY_RETRIES'].to_i : 3

    def initialize(options = {})
      @filter = options[:filter]
      @nodeset = options[:nodeset]
      @filter = {} unless (@filter || @nodeset)
    end

    def results
      raw_agent_inventory.collect { |node| node.results }
    end

    def ddl(agent_name)
      client_with_timeout(agent_name).ddl
    end

    def run(action, args)
      # Make sure args are symbol-keyed before passing to mcollective
      args.each do |k,v|
        if k.class != Symbol
          args[k.to_sym] = v
          args.delete k
        end
      end
      client = client_for(action)
      result = client.custom_request(action.name, args, filtered_nodes, { "identity" => node_filter })
      client.disconnect
      result
    end

    def client_with_timeout(client_name)
      client = rpcclient(client_name)
      client.discovery_timeout = DISCOVERY_TIMEOUT
      client
    end

    def filtered_nodes
      return nodeset if nodeset
      Emcee::Cache.new(filter.empty? ? 'empty_filter' : filter).data do
        Emcee.logger.debug("discovery is happening NOW!")
        client = filtered_client(client_with_timeout('rpcutil'))
        result = client.discover
        client.disconnect
        result
      end
    end

    def node_filter
      "/^(#{filtered_nodes.join('|')})$/"
    end

    def client_for(action)
      filtered_client(client_with_timeout(action.agent_name))
    end

    def filtered_client(client)
      if filter
        filter[:facts].each { |fact| client.fact_filter(*fact) } if filter[:facts]
        filter[:classes].each { |klass| client.class_filter(klass) } if filter[:classes]
        filter[:identities].each { |identity| client.identity_filter(identity) } if filter[:identities]
      elsif nodeset && !nodeset.empty?
        client.identity_filter(nodeset_to_identity(nodeset))
      end
      client
    end

    def raw_agent_inventory
      Emcee::Cache.new('raw_agent_inventory').data do
        Emcee.logger.debug("raw_agent_inventory is happening NOW!")
        results = []
        tries = 0
        while results.empty? && tries <= INVENTORY_RETRIES
          if tries > 0
            sleep(3)
            Emcee.logger.debug("TRYING AGAIN #{tries}")
          else
            Emcee.logger.debug("FIRST TRY")
          end
          client = client_with_timeout('rpcutil')
          results = client.agent_inventory
          client.disconnect
          tries += 1
        end
        raise "Agent Inventory failed; consider raising the LM_INVENTORY_RETRIES environment variable" if results.empty?
        results
      end
    end

    private

    def nodeset_to_identity(nodeset)
      escaped = nodeset.map { |n| Regexp.escape(n) }
      nodeset.empty? ? nil : "/^(#{escaped.join('|')})$/"
    end
  end
end
