require 'spec_helper'

describe CustomReportController do
  integrate_views

  def fixture(name)
    file = File.join(File.dirname(__FILE__), '..', 'fixtures', 'reports', "#{name}.yaml")
    return Report.create_from_yaml(File.read(file))
  end

  before :each do
    @foo_report = fixture('foo')
    @bar_report = fixture('bar')
    @baz_report = fixture('baz')
    @quux_report = fixture('quux_uncommon_changes')

    @group = NodeGroup.generate!
    @group.nodes = [@foo_report, @bar_report, @baz_report, @quux_report].map(&:node)
    @group.save!
  end

  describe 'without baseline data' do
    describe '#common' do
      it 'should redirect to the GroupReport page' do
        get :common, :id => @group.id
        response.should redirect_to(:controller => :group_report, :action => :common, :id => @group.id)
      end
    end

    describe '#individual' do
      it 'should redirect to the GroupReport page' do
        get :common, :id => @group.id
        response.should redirect_to(:controller => :group_report, :action => :common, :id => @group.id)
      end
    end

    describe '#export' do
      it "should not die when rendering a CSV" do
        get :export, :id => @group.id, :date => @foo_report.time.to_date.to_s, :format => 'csv'
        response.should be_success
      end
    end
  end

  describe 'with baseline data' do
    before :each do
      Summarizer.stubs(:delay).returns(Summarizer)

      foo_baseline_report = Factory(:inspect_report, :host => 'foo')
      bar_baseline_report = Factory(:inspect_report, :host => 'bar')
      baz_baseline_report = Factory(:inspect_report, :host => 'baz')

      @foo_baseline = NodeBaseline.create_from_report!(foo_baseline_report, @foo_report.time)
      @bar_baseline = NodeBaseline.create_from_report!(bar_baseline_report, @bar_report.time)
      @baz_baseline = NodeBaseline.create_from_report!(baz_baseline_report, @baz_report.time)

      Time.zone.stubs(:now).returns(@foo_report.time)

      Summarizer.compare_group_against_node(@group, @bar_report.node_id)
    end

    describe '(partial)' do
      before :each do
        NodeBaseline.find_by_node_id_and_reference_node_id(@foo_report.node_id, @bar_report.node_id).delete
      end

      describe '#common' do
        it 'should redirect to the GroupReport page' do
          get :common, :id => @group.id
          response.should redirect_to(:controller => :group_report, :action => :common, :id => @group.id)
        end
      end

      describe '#individual' do
        it 'should redirect to the GroupReport page' do
          get :common, :id => @group.id
          response.should redirect_to(:controller => :group_report, :action => :common, :id => @group.id)
        end
      end
    end

    describe '#export' do
      it "should not die when rendering a CSV" do
        get :export, :id => @group.id, :date => @foo_report.time.to_date.to_s, :format => 'csv'
        response.should be_success
      end
    end

    describe "#common" do
      it 'should assign useful common changes' do
        get :common, :id => @group.id
        assigns(:common_changes).should == [
          [
            "File[/etc/hosts]",
            [
              {
                :node_count => 1,
                :properties => {
                  :baseline => [],
                  :inspected => {"group"=>"0", "mode"=>"644", "owner"=>"0"}
                },
                :digest => "Pa1BjVFP4aRZxctihvTFjw==",
                :accepted => "no-action",
                :can_change => true,
                :baseline_count => 1
              },
              {
                :node_count => 1,
                :properties => {
                  :baseline => [],
                  :inspected => {"group"=>"0", "mode"=>"644", "owner"=>"1"}
                },
                :digest => "zpeS1VKyyK6YBhbl2qJm9g==",
                :accepted => "no-action",
                :can_change => true,
                :baseline_count => 1
              }
            ]
          ],
          [
            "File[/etc/sudoers]",
            [
              {
                :node_count => 1,
                :properties => {
                  :baseline => [],
                  :inspected => {"group"=>"0", "mode"=>"440", "owner"=>"0"}
                },
                :digest => "BdGCiBagW+SY+vZqKAvQdg==",
                :accepted => "no-action",
                :can_change => true,
                :baseline_count => 1
              },
              {
                :node_count => 1,
                :properties => {
                  :baseline => [],
                  :inspected => {"group"=>"0", "mode"=>"440", "owner"=>"1"}
                },
                :digest => "mJkVXDvihrvLFi/kL5AFVw==",
                :accepted => "no-action",
                :can_change => true,
                :baseline_count => 1
              }
            ]
          ],
          [
            "User[root]",
            [
              {
                :node_count => 1,
                :properties => {
                  :baseline => [],
                  :inspected => {"shell"=>"/bin/sh", "home"=>"/root"}
                },
                :digest => "0F4Kaeb6aW6gwIVfcz+zGA==",
                :accepted => "no-action",
                :can_change => true,
                :baseline_count => 1
              },
              {
                :node_count => 1,
                :properties => {
                  :baseline => [],
                  :inspected => {"shell"=>"/bin/sh", "home"=>"/var/root"}
                },
                :digest => "ny1vSG/0drWD9R1773TbNA==",
                :accepted => "no-action",
                :can_change => true,
                :baseline_count => 1
              }
            ]
          ]
        ]
      end
    end


    describe "#individual" do
      it 'should assign useful change counts' do
        get :individual, :id => @group.id
        assigns(:individual_changes).should == [
          OpenStruct.new(:node_id => @bar_report.node_id,  :node_name => 'bar',  :resource_differences => 0),
          OpenStruct.new(:node_id => @baz_report.node_id,  :node_name => 'baz',  :resource_differences => 3),
          OpenStruct.new(:node_id => @foo_report.node_id,  :node_name => 'foo',  :resource_differences => 3),
          OpenStruct.new(:node_id => @quux_report.node_id, :node_name => 'quux', :resource_differences => 2),
        ]
      end
    end
  end
end
